/*
 *    Virtual Steganographic Laboratory (VSL)
 *    Copyright (C) 2008-2011  M. Wegrzyn <bend-up@users.sourceforge.net>

 *    This file is part of Virtual Steganographic Laboratory (VSL).

 *    VSL is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.

 *    VSL is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.

 *    You should have received a copy of the GNU General Public License
 *    along with VSL.  If not, see <http://www.gnu.org/licenses/>.
 */
package pl.edu.zut.wi.vsl.app.gui.panels;

import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.KeyStroke;
import org.apache.log4j.Logger;
import pl.edu.zut.wi.vsl.app.Main;

/**
 * Panel or choosing output folder and name 
 * pattern for output file(s).
 * @author Michal Wegrzyn
 */
public class OutputPanel extends javax.swing.JPanel {

    /** Title for <code>JFileChooser</code> opened during 
     * choosing output folder for output files */
    public static final String PREFERRED_TITLE = "VSL - choose output folder";    
    private final static Logger logger = Logger.getLogger(OutputPanel.class);
    /** Directory in which files will be saved */
    private File selectedDirectory;

    /**
     * Creates new form OutputPanel.
     * @param dir selected output directory
     * @param pattern pattern of name for files that will be saved
     * @param format format of name for files that will be saved
     * @param showCombobox whether to show format combo box
     * @see pl.edu.zut.wi.vsl.app.modules.VslModule#outputDirectory
     */
    public OutputPanel(File dir, String pattern, String format,
            boolean showCombobox) {
        initComponents();
        selectedDirectory = dir;
        folderpathField.setText(dir.getAbsolutePath());
        namePatternField.setText(pattern);
        formatComboBox.setVisible(showCombobox);
        formatLabel.setVisible(showCombobox);
        formatComboBox.setSelectedItem(format);
        
        KeyStroke chooseKeyStroke = KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, 
                                                                    0, false);
        Action deleteAction = new AbstractAction() {
            //insert opens file chooser
            public void actionPerformed(ActionEvent e) {
                chooseDirectory();
            }
        };
        getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(chooseKeyStroke, 
                                                                    "INSERT");
        getActionMap().put("INSERT", deleteAction);
    }

    /**
     * Getter for output directory.
     * @return Path to directory in which files will be saved.
     */
    public File getDirectory() {
        return selectedDirectory;
    }

    /**
     * Getter for name pattern.
     * @return Pattern for filenames for output files.
     */
    public String getPattern() {
        return namePatternField.getText();
    }

    /**
     * Getter for name format.
     * @return Format for filenames for output files.
     */
    public String getFormat() {
        return formatComboBox.getSelectedItem().toString();
    }

    /**
     * Handles choosing output directory.
     */
    private void chooseDirectory() {
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        fileChooser.setDialogTitle(PREFERRED_TITLE);
        fileChooser.setFileHidingEnabled(false);
        fileChooser.setMultiSelectionEnabled(false);
        int returnValue = fileChooser.showDialog(getTopLevelAncestor(), 
                                                                    "Select");
        if (returnValue == JFileChooser.APPROVE_OPTION) {
            if (selectedDirectory != fileChooser.getSelectedFile()) {
                selectedDirectory = fileChooser.getSelectedFile();
                String dir = selectedDirectory.getAbsolutePath();
                logger.info("Selected output directory: " + dir);
                folderpathField.setText(dir);
            }
        }
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        folderpathLabel = new javax.swing.JLabel();
        folderpathField = new javax.swing.JTextField();
        chooseFolderButton = new javax.swing.JButton();
        namePatternLabel = new javax.swing.JLabel();
        namePatternField = new javax.swing.JTextField();
        formatLabel = new javax.swing.JLabel();
        formatComboBox = new javax.swing.JComboBox();

        folderpathLabel.setText("Folder:");
        folderpathLabel.setToolTipText("Folder for output file(s)");

        folderpathField.setEditable(false);
        folderpathField.setToolTipText("Path to the folder in which result file(s) should be be written");

        chooseFolderButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/pl/edu/zut/wi/vsl/app/gui/icons/folder22x22.png"))); // NOI18N
        chooseFolderButton.setToolTipText("Choose folder for output files (Ins)");
        chooseFolderButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                chooseFolderButtonActionPerformed(evt);
            }
        });

        namePatternLabel.setText("Name pattern:");
        namePatternLabel.setToolTipText("Name pattern for output files");

        namePatternField.setToolTipText("Specify output files name pattern (without fileformat)");

        formatLabel.setText("Format:");
        formatLabel.setToolTipText("Name pattern for output files");

        formatComboBox.setModel(new javax.swing.DefaultComboBoxModel(Main.getSupportedOutputFormats()));

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(namePatternLabel)
                    .addComponent(folderpathLabel)
                    .addComponent(formatLabel))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(formatComboBox, 0, 305, Short.MAX_VALUE)
                    .addComponent(namePatternField, javax.swing.GroupLayout.DEFAULT_SIZE, 305, Short.MAX_VALUE)
                    .addComponent(folderpathField, javax.swing.GroupLayout.DEFAULT_SIZE, 305, Short.MAX_VALUE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(chooseFolderButton)
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                        .addComponent(folderpathLabel)
                        .addComponent(folderpathField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addComponent(chooseFolderButton))
                .addGap(18, 18, 18)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(namePatternLabel)
                    .addComponent(namePatternField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(formatLabel)
                    .addComponent(formatComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addContainerGap(28, Short.MAX_VALUE))
        );
    }// </editor-fold>//GEN-END:initComponents

private void chooseFolderButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_chooseFolderButtonActionPerformed
    chooseDirectory();
}//GEN-LAST:event_chooseFolderButtonActionPerformed
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton chooseFolderButton;
    private javax.swing.JTextField folderpathField;
    private javax.swing.JLabel folderpathLabel;
    private javax.swing.JComboBox formatComboBox;
    private javax.swing.JLabel formatLabel;
    private javax.swing.JTextField namePatternField;
    private javax.swing.JLabel namePatternLabel;
    // End of variables declaration//GEN-END:variables
}
